@@ -14,6 +14,7 @@
 #include <boost/range/algorithm.hpp>
 #include <cmath>
 #include <fcitx-utils/log.h>
+#include <fcitx-utils/stringutils.h>
 
 namespace libime {
 
@@ -157,6 +158,22 @@ public:
         }
     }
 
+    void loadText(std::istream &in) {
+        clear();
+        std::string buf;
+        std::vector<std::string> lines;
+        while (std::getline(in, buf)) {
+            lines.emplace_back(buf);
+            if (lines.size() >= maxSize_)
+                break;
+        }
+        for (auto &line : lines | boost::adaptors::reversed) {
+            std::vector<std::string> sentence =
+                fcitx::stringutils::split(line, " ");
+            add(sentence);
+        }
+    }
+
     void save(std::ostream &out) {
         uint32_t count = recent_.size();
         throw_if_io_fail(marshall(out, count));
@@ -483,6 +500,11 @@ void HistoryBigram::load(std::istream &in) {
     }
 }
 
+void HistoryBigram::loadText(std::istream &in) {
+    FCITX_D();
+    boost::range::for_each(d->pools_, [&in](auto &pool) { pool.loadText(in); });
+}
+
 void HistoryBigram::save(std::ostream &out) {
     FCITX_D();
     throw_if_io_fail(marshall(out, historyBinaryFormatMagic));
